from __future__ import annotations as _annotations

from datetime import date, datetime, timedelta, timezone, tzinfo
from typing import TYPE_CHECKING, Any

from ._numeric import IsNumeric
from ._utils import Omit

if TYPE_CHECKING:
    from zoneinfo import ZoneInfo


class IsDatetime(IsNumeric[datetime]):
    """
    Check if the value is a datetime, and matches the given conditions.
    """

    allowed_types = datetime

    def __init__(
        self,
        *,
        approx: datetime | None = None,
        delta: timedelta | int | float | None = None,
        gt: datetime | None = None,
        lt: datetime | None = None,
        ge: datetime | None = None,
        le: datetime | None = None,
        unix_number: bool = False,
        iso_string: bool = False,
        format_string: str | None = None,
        enforce_tz: bool = True,
    ):
        """
        Args:
            approx: A value to approximately compare to.
            delta: The allowable different when comparing to the value to `approx`, if omitted 2 seconds is used,
                ints and floats are assumed to represent seconds and converted to `timedelta`s.
            gt: Value which the compared value should be greater than (after).
            lt: Value which the compared value should be less than (before).
            ge: Value which the compared value should be greater than (after) or equal to.
            le: Value which the compared value should be less than (before) or equal to.
            unix_number: whether to allow unix timestamp numbers in comparison
            iso_string: whether to allow iso formatted strings in comparison
            format_string: if provided, `format_string` is used with `datetime.strptime` to parse strings
            enforce_tz: whether timezone should be enforced in comparison, see below for more details

        Examples of basic usage:

        ```py title="IsDatetime"
        from datetime import datetime

        from dirty_equals import IsDatetime

        y2k = datetime(2000, 1, 1)
        assert datetime(2000, 1, 1) == IsDatetime(approx=y2k)
        # Note: this requires the system timezone to be UTC
        assert 946684800.123 == IsDatetime(approx=y2k, unix_number=True)
        assert datetime(2000, 1, 1, 0, 0, 9) == IsDatetime(approx=y2k, delta=10)
        assert '2000-01-01T00:00' == IsDatetime(approx=y2k, iso_string=True)

        assert datetime(2000, 1, 2) == IsDatetime(gt=y2k)
        assert datetime(1999, 1, 2) != IsDatetime(gt=y2k)
        ```
        """
        if isinstance(delta, (int, float)):
            delta = timedelta(seconds=delta)

        super().__init__(
            approx=approx,
            delta=delta,  # type: ignore[arg-type]
            gt=gt,
            lt=lt,
            ge=ge,
            le=le,
        )
        self.unix_number = unix_number
        self.iso_string = iso_string
        self.format_string = format_string
        self.enforce_tz = enforce_tz
        self._repr_kwargs.update(
            unix_number=Omit if unix_number is False else unix_number,
            iso_string=Omit if iso_string is False else iso_string,
            format_string=Omit if format_string is None else format_string,
            enforce_tz=Omit if enforce_tz is True else format_string,
        )

    def prepare(self, other: Any) -> datetime:
        if isinstance(other, datetime):
            dt = other
        elif isinstance(other, (float, int)):
            if self.unix_number:
                dt = datetime.fromtimestamp(other)
            else:
                raise TypeError('numbers not allowed')
        elif isinstance(other, str):
            if self.iso_string:
                dt = datetime.fromisoformat(other)
            elif self.format_string:
                dt = datetime.strptime(other, self.format_string)
            else:
                raise ValueError('not a valid datetime string')
        else:
            raise ValueError(f'{type(other)} not valid as datetime')

        if self.approx is not None and not self.enforce_tz and self.approx.tzinfo is None and dt.tzinfo is not None:
            dt = dt.replace(tzinfo=None)
        return dt

    def approx_equals(self, other: datetime, delta: timedelta) -> bool:
        if not super().approx_equals(other, delta):
            return False

        if self.enforce_tz:
            if self.approx.tzinfo is None:  # type: ignore[union-attr]
                return other.tzinfo is None
            else:
                approx_offset = self.approx.tzinfo.utcoffset(self.approx)  # type: ignore[union-attr]
                other_offset = other.tzinfo.utcoffset(other)  # type: ignore[union-attr]
                return approx_offset == other_offset
        else:
            return True


def _zoneinfo(tz: str) -> ZoneInfo:
    """
    Instantiate a `ZoneInfo` object from a string, falling back to `pytz.timezone` when `ZoneInfo` is not available
    (most likely on Python 3.8 and webassembly).
    """
    try:
        from zoneinfo import ZoneInfo
    except ImportError:
        try:
            import pytz
        except ImportError as e:
            raise ImportError('`pytz` or `zoneinfo` required for tz handling') from e
        else:
            return pytz.timezone(tz)  # type: ignore[return-value]
    else:
        return ZoneInfo(tz)


class IsNow(IsDatetime):
    """
    Check if a datetime is close to now, this is similar to `IsDatetime(approx=datetime.now())`,
    but slightly more powerful.
    """

    def __init__(
        self,
        *,
        delta: timedelta | int | float = 2,
        unix_number: bool = False,
        iso_string: bool = False,
        format_string: str | None = None,
        enforce_tz: bool = True,
        tz: str | tzinfo | None = None,
    ):
        """
        Args:
            delta: The allowable different when comparing to the value to now, if omitted 2 seconds is used,
                ints and floats are assumed to represent seconds and converted to `timedelta`s.
            unix_number: whether to allow unix timestamp numbers in comparison
            iso_string: whether to allow iso formatted strings in comparison
            format_string: if provided, `format_string` is used with `datetime.strptime` to parse strings
            enforce_tz: whether timezone should be enforced in comparison, see below for more details
            tz: either a `ZoneInfo`, a `datetime.timezone` or a string which will be passed to `ZoneInfo`,
                (or `pytz.timezone` on 3.8) to get a timezone,
                if provided now will be converted to this timezone.

        ```py title="IsNow"
        from datetime import datetime, timezone

        from dirty_equals import IsNow

        now = datetime.now()
        assert now == IsNow
        assert now.timestamp() == IsNow(unix_number=True)
        assert now.timestamp() != IsNow
        assert now.isoformat() == IsNow(iso_string=True)
        assert now.isoformat() != IsNow

        utc_now = datetime.utcnow().replace(tzinfo=timezone.utc)
        assert utc_now == IsNow(tz=timezone.utc)
        ```
        """
        if isinstance(tz, str):
            tz = _zoneinfo(tz)

        self.tz = tz

        approx = self._get_now()

        super().__init__(
            approx=approx,
            delta=delta,
            unix_number=unix_number,
            iso_string=iso_string,
            format_string=format_string,
            enforce_tz=enforce_tz,
        )
        if tz is not None:
            self._repr_kwargs['tz'] = tz

    def _get_now(self) -> datetime:
        if self.tz is None:
            return datetime.now()
        else:
            utc_now = datetime.now(tz=timezone.utc).replace(tzinfo=timezone.utc)
            return utc_now.astimezone(self.tz)

    def prepare(self, other: Any) -> datetime:
        # update approx for every comparing, to check if other value is dirty equal
        # to current moment of time
        self.approx = self._get_now()

        return super().prepare(other)


class IsDate(IsNumeric[date]):
    """
    Check if the value is a date, and matches the given conditions.
    """

    allowed_types = date

    def __init__(
        self,
        *,
        approx: date | None = None,
        delta: timedelta | int | float | None = None,
        gt: date | None = None,
        lt: date | None = None,
        ge: date | None = None,
        le: date | None = None,
        iso_string: bool = False,
        format_string: str | None = None,
    ):
        """
        Args:
            approx: A value to approximately compare to.
            delta: The allowable different when comparing to the value to now, if omitted 2 seconds is used,
                ints and floats are assumed to represent seconds and converted to `timedelta`s.
            gt: Value which the compared value should be greater than (after).
            lt: Value which the compared value should be less than (before).
            ge: Value which the compared value should be greater than (after) or equal to.
            le: Value which the compared value should be less than (before) or equal to.
            iso_string: whether to allow iso formatted strings in comparison
            format_string: if provided, `format_string` is used with `datetime.strptime` to parse strings

        Examples of basic usage:

        ```py title="IsDate"
        from datetime import date

        from dirty_equals import IsDate

        y2k = date(2000, 1, 1)
        assert date(2000, 1, 1) == IsDate(approx=y2k)
        assert '2000-01-01' == IsDate(approx=y2k, iso_string=True)

        assert date(2000, 1, 2) == IsDate(gt=y2k)
        assert date(1999, 1, 2) != IsDate(gt=y2k)
        ```
        """

        if delta is None:
            delta = timedelta()
        elif isinstance(delta, (int, float)):
            delta = timedelta(seconds=delta)

        super().__init__(approx=approx, gt=gt, lt=lt, ge=ge, le=le, delta=delta)  # type: ignore[arg-type]

        self.iso_string = iso_string
        self.format_string = format_string
        self._repr_kwargs.update(
            iso_string=Omit if iso_string is False else iso_string,
            format_string=Omit if format_string is None else format_string,
        )

    def prepare(self, other: Any) -> date:
        if type(other) is date:
            dt = other
        elif isinstance(other, str):
            if self.iso_string:
                dt = date.fromisoformat(other)
            elif self.format_string:
                dt = datetime.strptime(other, self.format_string).date()
            else:
                raise ValueError('not a valid date string')
        else:
            raise ValueError(f'{type(other)} not valid as date')

        return dt


class IsToday(IsDate):
    """
    Check if a date is today, this is similar to `IsDate(approx=date.today())`, but slightly more powerful.
    """

    def __init__(
        self,
        *,
        iso_string: bool = False,
        format_string: str | None = None,
    ):
        """
        Args:
            iso_string: whether to allow iso formatted strings in comparison
            format_string: if provided, `format_string` is used with `datetime.strptime` to parse strings
        ```py title="IsToday"
        from datetime import date, timedelta

        from dirty_equals import IsToday

        today = date.today()
        assert today == IsToday
        assert today.isoformat() == IsToday(iso_string=True)
        assert today.isoformat() != IsToday
        assert today + timedelta(days=1) != IsToday
        assert today.strftime('%Y/%m/%d') == IsToday(format_string='%Y/%m/%d')
        assert today.strftime('%Y/%m/%d') != IsToday()
        ```
        """

        super().__init__(approx=date.today(), iso_string=iso_string, format_string=format_string)



import math
from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import Any, Optional, Tuple, Type, TypeVar, Union

from ._base import DirtyEquals

__all__ = (
    'IsApprox',
    'IsNumeric',
    'IsNumber',
    'IsPositive',
    'IsNegative',
    'IsNonPositive',
    'IsNonNegative',
    'IsInt',
    'IsPositiveInt',
    'IsNegativeInt',
    'IsFloat',
    'IsPositiveFloat',
    'IsNegativeFloat',
)

from ._utils import Omit

AnyNumber = Union[int, float, Decimal]
N = TypeVar('N', int, float, Decimal, date, datetime, AnyNumber)


class IsNumeric(DirtyEquals[N]):
    """
    Base class for all numeric types, `IsNumeric` implements approximate and inequality comparisons,
    as well as the type checks.

    This class can be used directly or via any of its subclasses.
    """

    allowed_types: Union[Type[N], Tuple[type, ...]] = (int, float, Decimal, date, datetime)
    """It allows any of the types supported in its subclasses."""

    def __init__(
        self,
        *,
        exactly: Optional[N] = None,
        approx: Optional[N] = None,
        delta: Optional[N] = None,
        gt: Optional[N] = None,
        lt: Optional[N] = None,
        ge: Optional[N] = None,
        le: Optional[N] = None,
    ):
        """
        Args:
            exactly: A value to exactly compare to - useful when you want to make sure a value is an `int` or `float`,
                while also checking its value.
            approx: A value to approximately compare to.
            delta: The allowable different when comparing to the value to `approx`,
                if omitted `value / 100` is used except for datetimes where 2 seconds is used.
            gt: Value which the compared value should be greater than.
            lt: Value which the compared value should be less than.
            ge: Value which the compared value should be greater than or equal to.
            le: Value which the compared value should be less than or equal to.

        If not values are provided, only the type is checked.

        If `approx` is provided as well a `gt`, `lt`, `ge`, or `le`, a `TypeError` is raised.

        Example of direct usage:

        ```py title="IsNumeric"
        from datetime import datetime

        from dirty_equals import IsNumeric

        assert 1.0 == IsNumeric
        assert 4 == IsNumeric(gt=3)
        d = datetime(2020, 1, 1, 12, 0, 0)
        assert d == IsNumeric(approx=datetime(2020, 1, 1, 12, 0, 1))
        ```
        """
        self.exactly: Optional[N] = exactly
        if self.exactly is not None and (gt, lt, ge, le) != (None, None, None, None):
            raise TypeError('"exactly" cannot be combined with "gt", "lt", "ge", or "le"')
        if self.exactly is not None and approx is not None:
            raise TypeError('"exactly" cannot be combined with "approx"')
        self.approx: Optional[N] = approx
        if self.approx is not None and (gt, lt, ge, le) != (None, None, None, None):
            raise TypeError('"approx" cannot be combined with "gt", "lt", "ge", or "le"')
        self.delta: Optional[N] = delta
        self.gt: Optional[N] = gt
        self.lt: Optional[N] = lt
        self.ge: Optional[N] = ge
        self.le: Optional[N] = le
        self.has_bounds_checks = not all(f is None for f in (exactly, approx, delta, gt, lt, ge, le))
        kwargs = {
            'exactly': Omit if exactly is None else exactly,
            'approx': Omit if approx is None else approx,
            'delta': Omit if delta is None else delta,
            'gt': Omit if gt is None else gt,
            'lt': Omit if lt is None else lt,
            'ge': Omit if ge is None else ge,
            'le': Omit if le is None else le,
        }
        super().__init__(**kwargs)

    def prepare(self, other: Any) -> N:
        if other is True or other is False:
            raise TypeError('booleans are not numbers')
        elif not isinstance(other, self.allowed_types):
            raise TypeError(f'not a {self.allowed_types}')
        else:
            return other

    def equals(self, other: Any) -> bool:
        other = self.prepare(other)

        if self.has_bounds_checks:
            return self.bounds_checks(other)
        else:
            return True

    def bounds_checks(self, other: N) -> bool:
        if self.exactly is not None:
            return self.exactly == other
        elif self.approx is not None:
            if self.delta is None:
                if isinstance(other, date):
                    delta: Any = timedelta(seconds=2)
                else:
                    delta = abs(other / 100)
            else:
                delta = self.delta
            return self.approx_equals(other, delta)
        elif self.gt is not None and not other > self.gt:
            return False
        elif self.lt is not None and not other < self.lt:
            return False
        elif self.ge is not None and not other >= self.ge:
            return False
        elif self.le is not None and not other <= self.le:
            return False
        else:
            return True

    def approx_equals(self, other: Any, delta: Any) -> bool:
        return abs(self.approx - other) <= delta


class IsNumber(IsNumeric[AnyNumber]):
    """
    Base class for all types that can be used with all number types, e.g. numeric but not `date` or `datetime`.

    Inherits from [`IsNumeric`][dirty_equals.IsNumeric] and can therefore be initialised with any of its arguments.
    """

    allowed_types = int, float, Decimal
    """
    It allows any of the number types.
    """


Num = TypeVar('Num', int, float, Decimal)


class IsApprox(IsNumber):
    """
    Simplified subclass of [`IsNumber`][dirty_equals.IsNumber] that only allows approximate comparisons.
    """

    def __init__(self, approx: Num, *, delta: Optional[Num] = None):
        """
        Args:
            approx: A value to approximately compare to.
            delta: The allowable different when comparing to the value to `approx`, if omitted `value / 100` is used.

        ```py title="IsApprox"
        from dirty_equals import IsApprox

        assert 1.0 == IsApprox(1)
        assert 123 == IsApprox(120, delta=4)
        assert 201 == IsApprox(200)
        assert 201 != IsApprox(200, delta=0.1)
        ```
        """
        super().__init__(approx=approx, delta=delta)


class IsPositive(IsNumber):
    """
    Check that a value is positive (`> 0`), can be an `int`, a `float` or a `Decimal`
    (or indeed any value which implements `__gt__` for `0`).

    ```py title="IsPositive"
    from decimal import Decimal

    from dirty_equals import IsPositive

    assert 1.0 == IsPositive
    assert 1 == IsPositive
    assert Decimal('3.14') == IsPositive
    assert 0 != IsPositive
    assert -1 != IsPositive
    ```
    """

    def __init__(self) -> None:
        super().__init__(gt=0)
        self._repr_kwargs = {}


class IsNegative(IsNumber):
    """
    Check that a value is negative (`< 0`), can be an `int`, a `float` or a `Decimal`
    (or indeed any value which implements `__lt__` for `0`).

    ```py title="IsNegative"
    from decimal import Decimal

    from dirty_equals import IsNegative

    assert -1.0 == IsNegative
    assert -1 == IsNegative
    assert Decimal('-3.14') == IsNegative
    assert 0 != IsNegative
    assert 1 != IsNegative
    ```
    """

    def __init__(self) -> None:
        super().__init__(lt=0)
        self._repr_kwargs = {}


class IsNonNegative(IsNumber):
    """
    Check that a value is positive or zero (`>= 0`), can be an `int`, a `float` or a `Decimal`
    (or indeed any value which implements `__ge__` for `0`).

    ```py title="IsNonNegative"
    from decimal import Decimal

    from dirty_equals import IsNonNegative

    assert 1.0 == IsNonNegative
    assert 1 == IsNonNegative
    assert Decimal('3.14') == IsNonNegative
    assert 0 == IsNonNegative
    assert -1 != IsNonNegative
    assert Decimal('0') == IsNonNegative
    ```
    """

    def __init__(self) -> None:
        super().__init__(ge=0)
        self._repr_kwargs = {}


class IsNonPositive(IsNumber):
    """
    Check that a value is negative or zero (`<=0`), can be an `int`, a `float` or a `Decimal`
    (or indeed any value which implements `__le__` for `0`).

    ```py title="IsNonPositive"
    from decimal import Decimal

    from dirty_equals import IsNonPositive

    assert -1.0 == IsNonPositive
    assert -1 == IsNonPositive
    assert Decimal('-3.14') == IsNonPositive
    assert 0 == IsNonPositive
    assert 1 != IsNonPositive
    assert Decimal('-0') == IsNonPositive
    assert Decimal('0') == IsNonPositive
    ```
    """

    def __init__(self) -> None:
        super().__init__(le=0)
        self._repr_kwargs = {}


class IsInt(IsNumeric[int]):
    """
    Checks that a value is an integer.

    Inherits from [`IsNumeric`][dirty_equals.IsNumeric] and can therefore be initialised with any of its arguments.

    ```py title="IsInt"
    from dirty_equals import IsInt

    assert 1 == IsInt
    assert -2 == IsInt
    assert 1.0 != IsInt
    assert 'foobar' != IsInt
    assert True != IsInt
    assert 1 == IsInt(exactly=1)
    assert -2 != IsInt(exactly=1)
    ```
    """

    allowed_types = int
    """
    As the name suggests, only integers are allowed, booleans (`True` are `False`) are explicitly excluded although
    technically they are sub-types of `int`.
    """


class IsPositiveInt(IsInt):
    """
    Like [`IsPositive`][dirty_equals.IsPositive] but only for `int`s.

    ```py title="IsPositiveInt"
    from decimal import Decimal

    from dirty_equals import IsPositiveInt

    assert 1 == IsPositiveInt
    assert 1.0 != IsPositiveInt
    assert Decimal('3.14') != IsPositiveInt
    assert 0 != IsPositiveInt
    assert -1 != IsPositiveInt
    ```
    """

    def __init__(self) -> None:
        super().__init__(gt=0)
        self._repr_kwargs = {}


class IsNegativeInt(IsInt):
    """
    Like [`IsNegative`][dirty_equals.IsNegative] but only for `int`s.

    ```py title="IsNegativeInt"
    from decimal import Decimal

    from dirty_equals import IsNegativeInt

    assert -1 == IsNegativeInt
    assert -1.0 != IsNegativeInt
    assert Decimal('-3.14') != IsNegativeInt
    assert 0 != IsNegativeInt
    assert 1 != IsNegativeInt
    ```
    """

    def __init__(self) -> None:
        super().__init__(lt=0)
        self._repr_kwargs = {}


class IsFloat(IsNumeric[float]):
    """
    Checks that a value is a float.

    Inherits from [`IsNumeric`][dirty_equals.IsNumeric] and can therefore be initialised with any of its arguments.

    ```py title="IsFloat"
    from dirty_equals import IsFloat

    assert 1.0 == IsFloat
    assert 1 != IsFloat
    assert 1.0 == IsFloat(exactly=1.0)
    assert 1.001 != IsFloat(exactly=1.0)
    ```
    """

    allowed_types = float
    """
    As the name suggests, only floats are allowed.
    """


class IsPositiveFloat(IsFloat):
    """
    Like [`IsPositive`][dirty_equals.IsPositive] but only for `float`s.

    ```py title="IsPositiveFloat"
    from decimal import Decimal

    from dirty_equals import IsPositiveFloat

    assert 1.0 == IsPositiveFloat
    assert 1 != IsPositiveFloat
    assert Decimal('3.14') != IsPositiveFloat
    assert 0.0 != IsPositiveFloat
    assert -1.0 != IsPositiveFloat
    ```
    """

    def __init__(self) -> None:
        super().__init__(gt=0)
        self._repr_kwargs = {}


class IsNegativeFloat(IsFloat):
    """
    Like [`IsNegative`][dirty_equals.IsNegative] but only for `float`s.

    ```py title="IsNegativeFloat"
    from decimal import Decimal

    from dirty_equals import IsNegativeFloat

    assert -1.0 == IsNegativeFloat
    assert -1 != IsNegativeFloat
    assert Decimal('-3.14') != IsNegativeFloat
    assert 0.0 != IsNegativeFloat
    assert 1.0 != IsNegativeFloat
    ```
    """

    def __init__(self) -> None:
        super().__init__(lt=0)
        self._repr_kwargs = {}


class IsFloatInf(IsFloat):
    """
    Checks that a value is float and infinite (positive or negative).

    Inherits from [`IsFloat`][dirty_equals.IsFloat].

    ```py title="IsFloatInf"
    from dirty_equals import IsFloatInf

    assert float('inf') == IsFloatInf
    assert float('-inf') == IsFloatInf
    assert 1.0 != IsFloatInf
    ```
    """

    def equals(self, other: Any) -> bool:
        other = self.prepare(other)
        return math.isinf(other)


class IsFloatInfPos(IsFloatInf):
    """
    Checks that a value is float and positive infinite.

    Inherits from [`IsFloatInf`][dirty_equals.IsFloatInf].

    ```py title="IsFloatInfPos"
    from dirty_equals import IsFloatInfPos

    assert float('inf') == IsFloatInfPos
    assert -float('-inf') == IsFloatInfPos
    assert -float('inf') != IsFloatInfPos
    assert float('-inf') != IsFloatInfPos
    ```
    """

    def __init__(self) -> None:
        super().__init__(gt=0)
        self._repr_kwargs = {}

    def equals(self, other: Any) -> bool:
        return self.bounds_checks(other) and super().equals(other)


class IsFloatInfNeg(IsFloatInf):
    """
    Checks that a value is float and negative infinite.

    Inherits from [`IsFloatInf`][dirty_equals.IsFloatInf].

    ```py title="IsFloatInfNeg"
    from dirty_equals import IsFloatInfNeg

    assert -float('inf') == IsFloatInfNeg
    assert float('-inf') == IsFloatInfNeg
    assert float('inf') != IsFloatInfNeg
    assert -float('-inf') != IsFloatInfNeg
    ```
    """

    def __init__(self) -> None:
        super().__init__(lt=0)
        self._repr_kwargs = {}

    def equals(self, other: Any) -> bool:
        return self.bounds_checks(other) and super().equals(other)


class IsFloatNan(IsFloat):
    """
    Checks that a value is float and nan (not a number).

    Inherits from [`IsFloat`][dirty_equals.IsFloat].

    ```py title="IsFloatNan"
    from dirty_equals import IsFloatNan

    assert float('nan') == IsFloatNan
    assert 1.0 != IsFloatNan
    ```
    """

    def equals(self, other: Any) -> bool:
        other = self.prepare(other)
        return math.isnan(other)



from __future__ import annotations

import json
import re
from dataclasses import asdict, is_dataclass
from enum import Enum
from functools import lru_cache
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network, ip_network
from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar, Union, overload
from uuid import UUID

from ._base import DirtyEquals
from ._dict import IsDict
from ._utils import Omit, plain_repr

if TYPE_CHECKING:
    from pydantic import TypeAdapter


class IsUUID(DirtyEquals[UUID]):
    """
    A class that checks if a value is a valid UUID, optionally checking UUID version.
    """

    def __init__(self, version: Literal[None, 1, 2, 3, 4, 5] = None):
        """
        Args:
            version: The version of the UUID to check, if omitted, all versions are accepted.

        ```py title="IsUUID"
        import uuid

        from dirty_equals import IsUUID

        assert 'edf9f29e-45c7-431c-99db-28ea44df9785' == IsUUID
        assert 'edf9f29e-45c7-431c-99db-28ea44df9785' == IsUUID(4)
        assert 'edf9f29e45c7431c99db28ea44df9785' == IsUUID(4)
        assert 'edf9f29e-45c7-431c-99db-28ea44df9785' != IsUUID(5)
        assert uuid.uuid4() == IsUUID(4)
        ```
        """
        self.version = version
        super().__init__(version or plain_repr('*'))

    def equals(self, other: Any) -> bool:
        if isinstance(other, UUID):
            uuid = other
        elif isinstance(other, str):
            uuid = UUID(other)
            if self.version is not None and uuid.version != self.version:
                return False
        else:
            return False

        if self.version:
            return uuid.version == self.version
        else:
            return True


AnyJson = object
JsonType = TypeVar('JsonType', AnyJson, Any)


class IsJson(DirtyEquals[JsonType]):
    """
    A class that checks if a value is a JSON object, and check the contents of the JSON.
    """

    @overload
    def __init__(self, expected_value: JsonType = AnyJson): ...

    @overload
    def __init__(self, **expected_kwargs: Any): ...

    def __init__(self, expected_value: JsonType = AnyJson, **expected_kwargs: Any):
        """
        Args:
            expected_value: Value to compare the JSON to, if omitted, any JSON is accepted.
            **expected_kwargs: Keyword arguments forming a dict to compare the JSON to,
                `expected_value` and `expected_kwargs` may not be combined.

        As with any `dirty_equals` type, types can be nested to provide more complex checks.

        !!! note
            Like [`IsInstance`][dirty_equals.IsInstance], `IsJson` can be parameterized or initialised with a value -
            `IsJson[xyz]` is exactly equivalent to `IsJson(xyz)`.

            This allows usage to be analogous to type hints.


        ```py title="IsJson"
        from dirty_equals import IsJson, IsPositiveInt, IsStrictDict

        assert '{"a": 1, "b": 2}' == IsJson
        assert '{"a": 1, "b": 2}' == IsJson(a=1, b=2)
        assert '{"a": 1}' != IsJson(a=2)
        assert 'invalid json' != IsJson
        assert '{"a": 1}' == IsJson(a=IsPositiveInt)
        assert '"just a quoted string"' == IsJson('just a quoted string')

        assert '{"a": 1, "b": 2}' == IsJson[IsStrictDict(a=1, b=2)]
        assert '{"b": 2, "a": 1}' != IsJson[IsStrictDict(a=1, b=2)]
        ```
        """
        if expected_kwargs:
            if expected_value is not AnyJson:
                raise TypeError('IsJson requires either an argument or kwargs, not both')
            self.expected_value: Any = expected_kwargs
        else:
            self.expected_value = expected_value
        super().__init__(plain_repr('*') if expected_value is AnyJson else expected_value)

    def __class_getitem__(cls, expected_type: JsonType) -> IsJson[JsonType]:
        return cls(expected_type)

    def equals(self, other: Any) -> bool:
        if isinstance(other, (str, bytes)):
            v = json.loads(other)
            if self.expected_value is AnyJson:
                return True
            else:
                return v == self.expected_value
        else:
            return False


class FunctionCheck(DirtyEquals[Any]):
    """
    Use a function to check if a value "equals" whatever you want to check
    """

    def __init__(self, func: Callable[[Any], bool]):
        """
        Args:
            func: callable that takes a value and returns a bool.

        ```py title="FunctionCheck"
        from dirty_equals import FunctionCheck

        def is_even(x):
            return x % 2 == 0

        assert 2 == FunctionCheck(is_even)
        assert 3 != FunctionCheck(is_even)
        ```
        """
        self.func = func
        super().__init__(plain_repr(func.__name__))

    def equals(self, other: Any) -> bool:
        return self.func(other)


T = TypeVar('T')


@lru_cache
def _build_type_adapter(ta: type[TypeAdapter[T]], schema: T) -> TypeAdapter[T]:
    return ta(schema)


_allowed_url_attribute_checks: set[str] = {
    'scheme',
    'host',
    'host_type',
    'user',
    'password',
    'port',
    'path',
    'query',
    'fragment',
}


class IsUrl(DirtyEquals[Any]):
    """
    A class that checks if a value is a valid URL, optionally checking different URL types and attributes with
    [Pydantic](https://pydantic-docs.helpmanual.io/usage/types/#urls).
    """

    def __init__(
        self,
        any_url: bool = False,
        any_http_url: bool = False,
        http_url: bool = False,
        file_url: bool = False,
        postgres_dsn: bool = False,
        ampqp_dsn: bool = False,
        redis_dsn: bool = False,
        **expected_attributes: Any,
    ):
        """
        Args:
            any_url: any scheme allowed, host required
            any_http_url: scheme http or https, host required
            http_url: scheme http or https, host required, max length 2083
            file_url: scheme file, host not required
            postgres_dsn: user info required
            ampqp_dsn: schema amqp or amqps, user info not required, host not required
            redis_dsn: scheme redis or rediss, user info not required, host not required
            **expected_attributes: Expected values for url attributes
        ```py title="IsUrl"
        from dirty_equals import IsUrl

        assert 'https://example.com' == IsUrl
        assert 'https://example.com' == IsUrl(host='example.com')
        assert 'https://example.com' == IsUrl(scheme='https')
        assert 'https://example.com' != IsUrl(scheme='http')
        assert 'postgres://user:pass@localhost:5432/app' == IsUrl(postgres_dsn=True)
        assert 'postgres://user:pass@localhost:5432/app' != IsUrl(http_url=True)
        ```
        """
        try:
            from pydantic import (
                AmqpDsn,
                AnyHttpUrl,
                AnyUrl,
                FileUrl,
                HttpUrl,
                PostgresDsn,
                RedisDsn,
                TypeAdapter,
                ValidationError,
            )

            self.ValidationError = ValidationError
        except ImportError as e:  # pragma: no cover
            raise ImportError('Pydantic V2 is not installed, run `pip install dirty-equals[pydantic]`') from e
        url_type_mappings = {
            AnyUrl: any_url,
            AnyHttpUrl: any_http_url,
            HttpUrl: http_url,
            FileUrl: file_url,
            PostgresDsn: postgres_dsn,
            AmqpDsn: ampqp_dsn,
            RedisDsn: redis_dsn,
        }
        url_types_sum = sum(url_type_mappings.values())
        if url_types_sum == 0:
            url_type: Any = AnyUrl
        elif url_types_sum == 1:
            url_type = max(url_type_mappings, key=url_type_mappings.get)  # type: ignore[arg-type]
        else:
            raise ValueError('You can only check against one Pydantic url type at a time')

        self.type_adapter = _build_type_adapter(TypeAdapter, url_type)

        for item in expected_attributes:
            if item not in _allowed_url_attribute_checks:
                raise TypeError(
                    'IsURL only checks these attributes: scheme, host, host_type, user, password, '
                    'port, path, query, fragment'
                )
        self.attribute_checks = expected_attributes
        super().__init__()

    def equals(self, other: Any) -> bool:
        try:
            other_url = self.type_adapter.validate_python(other)
        except self.ValidationError:
            raise ValueError('Invalid URL')

        # we now check that str() of the parsed URL equals its original value
        # so that invalid encodings fail
        # we remove trailing slashes since they're added by pydantic's URL parsing, but don't mean `other` is invalid
        other_url_str = str(other_url)
        if not other.endswith('/') and other_url_str.endswith('/'):
            other_url_str = other_url_str[:-1]
        equal = other_url_str == other

        if not self.attribute_checks:
            return equal

        for attribute, expected in self.attribute_checks.items():
            if getattr(other_url, attribute) != expected:
                return False
        return equal


HashTypes = Literal['md5', 'sha-1', 'sha-256']


class IsHash(DirtyEquals[str]):
    """
    A class that checks if a value is a valid common hash type, using a simple length and allowed characters regex.
    """

    def __init__(self, hash_type: HashTypes):
        """
        Args:
            hash_type: The hash type to check. Must be specified.

        ```py title="IsHash"
        from dirty_equals import IsHash

        assert 'f1e069787ece74531d112559945c6871' == IsHash('md5')
        assert b'f1e069787ece74531d112559945c6871' == IsHash('md5')
        assert 'f1e069787ece74531d112559945c6871' != IsHash('sha-256')
        assert 'F1E069787ECE74531D112559945C6871' == IsHash('md5')
        assert '40bd001563085fc35165329ea1ff5c5ecbdbbeef' == IsHash('sha-1')
        assert 'a665a45920422f9d417e4867efdc4fb8a04a1f3fff1fa07e998e86f7f7a27ae3' == IsHash(
            'sha-256'
        )
        ```
        """

        allowed_hashes = HashTypes.__args__  # type: ignore[attr-defined]
        if hash_type not in allowed_hashes:
            raise ValueError(f"Hash type must be one of the following values: {', '.join(allowed_hashes)}")

        self.hash_type = hash_type
        super().__init__(hash_type)

    def equals(self, other: Any) -> bool:
        if isinstance(other, str):
            s = other
        elif isinstance(other, (bytes, bytearray)):
            s = other.decode()
        else:
            return False
        hash_type_regex_patterns = {
            'md5': r'[a-fA-F\d]{32}',
            'sha-1': r'[a-fA-F\d]{40}',
            'sha-256': r'[a-fA-F\d]{64}',
        }
        return bool(re.fullmatch(hash_type_regex_patterns[self.hash_type], s))


IP = TypeVar('IP', IPv4Address, IPv4Network, IPv6Address, IPv6Network, Union[str, int, bytes])


class IsIP(DirtyEquals[IP]):
    """
    A class that checks if a value is a valid IP address, optionally checking IP version, netmask.
    """

    def __init__(self, *, version: Literal[None, 4, 6] = None, netmask: str | None = None):
        """
        Args:
            version: The version of the IP to check, if omitted, versions 4 and 6 are both accepted.
            netmask: The netmask of the IP to check, if omitted, any netmask is accepted. Requires version.

        ```py title="IsIP"
        from ipaddress import IPv4Address, IPv4Network, IPv6Address

        from dirty_equals import IsIP

        assert '179.27.154.96' == IsIP
        assert '179.27.154.96' == IsIP(version=4)
        assert '2001:0db8:0a0b:12f0:0000:0000:0000:0001' == IsIP(version=6)
        assert IPv4Address('127.0.0.1') == IsIP
        assert IPv4Network('43.48.0.0/12') == IsIP
        assert IPv6Address('::eeff:ae3f:d473') == IsIP
        assert '54.43.53.219/10' == IsIP(version=4, netmask='255.192.0.0')
        assert '54.43.53.219/10' == IsIP(version=4, netmask=4290772992)
        assert '::ffff:aebf:d473/12' == IsIP(version=6, netmask='fff0::')
        assert 3232235521 == IsIP
        ```
        """
        self.version = version
        if netmask and not self.version:
            raise TypeError('To check the netmask you must specify the IP version')
        self.netmask = netmask
        super().__init__(version=version or Omit, netmask=netmask or Omit)

    def equals(self, other: Any) -> bool:
        if isinstance(other, (IPv4Network, IPv6Network)):
            ip = other
        elif isinstance(other, (str, bytes, int, IPv4Address, IPv6Address)):
            ip = ip_network(other, strict=False)
        else:
            return False

        if self.version:
            if self.netmask:
                version_check = self.version == ip.version
                address_format = {4: IPv4Address, 6: IPv6Address}[self.version]
                netmask_check = int(address_format(self.netmask)) == int(ip.netmask)
                return version_check and netmask_check
            elif self.version != ip.version:
                return False

        return True


class IsDataclassType(DirtyEquals[Any]):
    """
    Checks that an object is a dataclass type.

    Inherits from [`DirtyEquals`][dirty_equals.DirtyEquals].

    ```py title="IsDataclassType"
    from dataclasses import dataclass
    from dirty_equals import IsDataclassType

    @dataclass
    class Foo:
        a: int
        b: int

    foo = Foo(1, 2)

    assert Foo == IsDataclassType
    assert foo != IsDataclassType
    ```
    """

    def equals(self, other: Any) -> bool:
        return is_dataclass(other) and isinstance(other, type)


class IsDataclass(DirtyEquals[Any]):
    """
    Checks that an object is an instance of a dataclass.

    Inherits from [`DirtyEquals`][dirty_equals.DirtyEquals] and it can be initialised with specific keyword arguments to
    check exactness of dataclass fields, by comparing the instance `__dict__`  with [`IsDict`][dirty_equals.IsDict].
    Moreover it is possible to check for strictness and partialness of the dataclass, by setting the `strict` and
    `partial` attributes using the `.settings(strict=..., partial=...)` method.

    Remark that passing no kwargs to `IsDataclass` initialization means fields are not checked, not that the dataclass
    is empty, namely `IsDataclass()` is the same as `IsDataclass`.

    ```py title="IsDataclass"
    from dataclasses import dataclass
    from dirty_equals import IsInt, IsDataclass

    @dataclass
    class Foo:
        a: int
        b: int
        c: str

    foo = Foo(1, 2, 'c')

    assert foo == IsDataclass
    assert foo == IsDataclass(a=IsInt, b=2, c='c')
    assert foo == IsDataclass(b=2, a=1).settings(partial=True)
    assert foo != IsDataclass(a=IsInt, b=2).settings(strict=True)
    assert foo == IsDataclass(a=IsInt, b=2).settings(strict=True, partial=True)
    assert foo != IsDataclass(b=2, a=1).settings(strict=True, partial=True)
    ```
    """

    def __init__(self, **fields: Any):
        """
        Args:
            fields: key-value pairs of field-value to check for.
        """
        self.strict = False
        self.partial = False
        self._post_init()
        super().__init__(**fields)

    def _post_init(self) -> None:
        pass

    def equals(self, other: Any) -> bool:
        if is_dataclass(other) and not isinstance(other, type):
            if self._repr_kwargs:
                return self._fields_check(other)
            else:
                return True
        else:
            return False

    def settings(
        self,
        *,
        strict: bool | None = None,
        partial: bool | None = None,
    ) -> IsDataclass:
        """Allows to customise the behaviour of `IsDataclass`, technically a new `IsDataclass` to allow chaining."""
        new_cls = self.__class__(**self._repr_kwargs)
        new_cls.__dict__ = self.__dict__.copy()

        if strict is not None:
            new_cls.strict = strict
        if partial is not None:
            new_cls.partial = partial

        return new_cls

    def _fields_check(self, other: Any) -> bool:
        """
        Checks exactness of fields using [`IsDict`][dirty_equals.IsDict] with given settings.

        Remark that if this method is called, then `other` is an instance of a dataclass, therefore we can call
        `dataclasses.asdict` to convert to a dict.
        """
        return asdict(other) == IsDict(self._repr_kwargs).settings(strict=self.strict, partial=self.partial)


class IsPartialDataclass(IsDataclass):
    """
    Inherits from [`IsDataclass`][dirty_equals.IsDataclass] with `partial=True` by default.

    ```py title="IsPartialDataclass"
    from dataclasses import dataclass
    from dirty_equals import IsInt, IsPartialDataclass

    @dataclass
    class Foo:
        a: int
        b: int
        c: str = 'c'

    foo = Foo(1, 2, 'c')

    assert foo == IsPartialDataclass
    assert foo == IsPartialDataclass(a=1)
    assert foo == IsPartialDataclass(b=2, a=IsInt)
    assert foo != IsPartialDataclass(b=2, a=IsInt).settings(strict=True)
    assert Foo != IsPartialDataclass
    ```
    """

    def _post_init(self) -> None:
        self.partial = True


class IsStrictDataclass(IsDataclass):
    """
    Inherits from [`IsDataclass`][dirty_equals.IsDataclass] with `strict=True` by default.

    ```py title="IsStrictDataclass"
    from dataclasses import dataclass
    from dirty_equals import IsInt, IsStrictDataclass

    @dataclass
    class Foo:
        a: int
        b: int
        c: str = 'c'

    foo = Foo(1, 2, 'c')

    assert foo == IsStrictDataclass
    assert foo == IsStrictDataclass(
        a=IsInt,
        b=2,
    ).settings(partial=True)
    assert foo != IsStrictDataclass(
        a=IsInt,
        b=2,
    ).settings(partial=False)
    assert foo != IsStrictDataclass(b=2, a=IsInt, c='c')
    ```
    """

    def _post_init(self) -> None:
        self.strict = True


class IsEnum(DirtyEquals[Enum]):
    """
    Checks if an instance is an Enum.

    Inherits from [`DirtyEquals`][dirty_equals.DirtyEquals].

    ```py title="IsEnum"
    from enum import Enum, auto
    from dirty_equals import IsEnum

    class ExampleEnum(Enum):
        a = auto()
        b = auto()

    a = ExampleEnum.a
    assert a == IsEnum
    assert a == IsEnum(ExampleEnum)
    ```
    """

    def __init__(self, enum_cls: type[Enum] = Enum):
        """
        Args:
            enum_cls: Enum class to check against.
        """
        self._enum_cls = enum_cls
        self._enum_values = {i.value for i in enum_cls}

    def equals(self, other: Any) -> bool:
        if isinstance(other, Enum):
            return isinstance(other, self._enum_cls)
        else:
            return other in self._enum_values



import io
from abc import ABCMeta
from pprint import PrettyPrinter
from typing import TYPE_CHECKING, Any, Dict, Generic, Iterable, Optional, Protocol, Tuple, TypeVar

from ._utils import Omit

if TYPE_CHECKING:
    from typing import TypeAlias, Union  # noqa: F401

__all__ = 'DirtyEqualsMeta', 'DirtyEquals', 'AnyThing', 'IsOneOf'


class DirtyEqualsMeta(ABCMeta):
    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True

        # this is required as fancy things happen when creating generics which include equals checks, without it,
        # we get some recursive errors
        if self is DirtyEquals or other is Generic or other is Protocol:
            return False
        else:
            try:
                return self() == other
            except TypeError:
                # we don't want to raise a type error here since somewhere deep in pytest it does something like
                # type(a) == type(b), if we raised TypeError we would upset the pytest error message
                return False

    def __or__(self, other: Any) -> 'DirtyOr':  # type: ignore[override]
        return DirtyOr(self, other)

    def __and__(self, other: Any) -> 'DirtyAnd':
        return DirtyAnd(self, other)

    def __invert__(self) -> 'DirtyNot':
        return DirtyNot(self)

    def __hash__(self) -> int:
        return hash(self.__name__)

    def __repr__(self) -> str:
        return self.__name__


T = TypeVar('T')


class DirtyEquals(Generic[T], metaclass=DirtyEqualsMeta):
    """
    Base type for all *dirty-equals* types.
    """

    __slots__ = '_other', '_was_equal', '_repr_args', '_repr_kwargs'

    def __init__(self, *repr_args: Any, **repr_kwargs: Any):
        """
        Args:
            *repr_args: unnamed args to be used in `__repr__`
            **repr_kwargs: named args to be used in `__repr__`
        """
        self._other: Any = None
        self._was_equal: Optional[bool] = None
        self._repr_args: Iterable[Any] = repr_args
        self._repr_kwargs: Dict[str, Any] = repr_kwargs

    def equals(self, other: Any) -> bool:
        """
        Abstract method, must be implemented by subclasses.

        `TypeError` and `ValueError` are caught in `__eq__` and indicate `other` is not equals to this type.
        """
        raise NotImplementedError()

    @property
    def value(self) -> T:
        """
        Property to get the value last successfully compared to this object.

        This is seldom very useful, put it's provided for completeness.

        Example of usage:

        ```py title=".values"
        from dirty_equals import IsStr

        token_is_str = IsStr(regex=r't-.+')
        assert 't-123' == token_is_str

        print(token_is_str.value)
        #> t-123
        ```
        """
        if self._was_equal:
            return self._other
        else:
            raise AttributeError('value is not available until __eq__ has been called')

    def __eq__(self, other: Any) -> bool:
        self._other = other
        try:
            self._was_equal = self.equals(other)
        except (TypeError, ValueError):
            self._was_equal = False

        return self._was_equal

    def __ne__(self, other: Any) -> bool:
        # We don't set _was_equal to avoid strange errors in pytest
        self._other = other
        try:
            return not self.equals(other)
        except (TypeError, ValueError):
            return True

    def __or__(self, other: Any) -> 'DirtyOr':
        return DirtyOr(self, other)

    def __and__(self, other: Any) -> 'DirtyAnd':
        return DirtyAnd(self, other)

    def __invert__(self) -> 'DirtyNot':
        return DirtyNot(self)

    def _repr_ne(self) -> str:
        args = [repr(arg) for arg in self._repr_args if arg is not Omit]
        args += [f'{k}={v!r}' for k, v in self._repr_kwargs.items() if v is not Omit]
        return f'{self.__class__.__name__}({", ".join(args)})'

    def __repr__(self) -> str:
        if self._was_equal:
            # if we've got the correct value return it to aid in diffs
            return repr(self._other)
        else:
            # else return something which explains what's going on.
            return self._repr_ne()

    def _pprint_format(self, pprinter: PrettyPrinter, stream: io.StringIO, *args: Any, **kwargs: Any) -> None:
        # pytest diffs use pprint to format objects, so we patch pprint to call this method
        # for DirtyEquals objects. So this method needs to follow the same pattern as __repr__.
        # We check that the protected _format method actually exists
        # to be safe and to make linters happy.
        if self._was_equal and hasattr(pprinter, '_format'):
            pprinter._format(self._other, stream, *args, **kwargs)
        else:
            stream.write(repr(self))  # i.e. self._repr_ne() (for now)


# Patch pprint to call _pprint_format for DirtyEquals objects
# Check that the protected attribute _dispatch exists to be safe and to make linters happy.
# The reason we modify _dispatch rather than _format
# is that pytest sometimes uses a subclass of PrettyPrinter which overrides _format.
if hasattr(PrettyPrinter, '_dispatch'):  # pragma: no branch
    PrettyPrinter._dispatch[DirtyEquals.__repr__] = lambda pprinter, obj, *args, **kwargs: obj._pprint_format(
        pprinter, *args, **kwargs
    )


InstanceOrType: 'TypeAlias' = 'Union[DirtyEquals[Any], DirtyEqualsMeta]'


class DirtyOr(DirtyEquals[Any]):
    def __init__(self, a: 'InstanceOrType', b: 'InstanceOrType', *extra: 'InstanceOrType'):
        self.dirties = (a, b) + extra
        super().__init__()

    def equals(self, other: Any) -> bool:
        return any(d == other for d in self.dirties)

    def _repr_ne(self) -> str:
        return ' | '.join(_repr_ne(d) for d in self.dirties)


class DirtyAnd(DirtyEquals[Any]):
    def __init__(self, a: InstanceOrType, b: InstanceOrType, *extra: InstanceOrType):
        self.dirties = (a, b) + extra
        super().__init__()

    def equals(self, other: Any) -> bool:
        return all(d == other for d in self.dirties)

    def _repr_ne(self) -> str:
        return ' & '.join(_repr_ne(d) for d in self.dirties)


class DirtyNot(DirtyEquals[Any]):
    def __init__(self, subject: InstanceOrType):
        self.subject = subject
        super().__init__()

    def equals(self, other: Any) -> bool:
        return self.subject != other

    def _repr_ne(self) -> str:
        return f'~{_repr_ne(self.subject)}'


def _repr_ne(v: InstanceOrType) -> str:
    if isinstance(v, DirtyEqualsMeta):
        return repr(v)
    else:
        return v._repr_ne()


class AnyThing(DirtyEquals[Any]):
    """
    A type which matches any value. `AnyThing` isn't generally very useful on its own, but can be used within
    other comparisons.

    ```py title="AnyThing"
    from dirty_equals import AnyThing, IsList, IsStrictDict

    assert 1 == AnyThing
    assert 'foobar' == AnyThing
    assert [1, 2, 3] == AnyThing

    assert [1, 2, 3] == IsList(AnyThing, 2, 3)

    assert {'a': 1, 'b': 2, 'c': 3} == IsStrictDict(a=1, b=AnyThing, c=3)
    ```
    """

    def equals(self, other: Any) -> bool:
        return True


class IsOneOf(DirtyEquals[Any]):
    """
    A type which checks that the value is equal to one of the given values.

    Can be useful with boolean operators.
    """

    def __init__(self, expected_value: Any, *more_expected_values: Any):
        """
        Args:
            expected_value: Expected value for equals to return true.
            *more_expected_values: More expected values for equals to return true.

        ```py title="IsOneOf"
        from dirty_equals import Contains, IsOneOf

        assert 1 == IsOneOf(1, 2, 3)
        assert 4 != IsOneOf(1, 2, 3)
        # check that a list either contain 1 or is empty
        assert [1, 2, 3] == Contains(1) | IsOneOf([])
        assert [] == Contains(1) | IsOneOf([])
        ```
        """
        self.expected_values: Tuple[Any, ...] = (expected_value,) + more_expected_values
        super().__init__(*self.expected_values)

    def equals(self, other: Any) -> bool:
        return any(other == e for e in self.expected_values)
